/*
								+---------------------------------+
								|                                 |
								|   ***   Font tex-atlas   ***    |
								|                                 |
								|  Copyright   -tHE SWINe- 2007  |
								|                                 |
								|          FontTexture.h          |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 2007-06-04
 *
 *	fixed clipped characters artifacts occuring on some fonts with strong character
 *	kerning by rendering glyphs into the center of the bitmap, thus ignoring character metrics
 *
 *	note MSVC compiler has strange error where it incorrectly reports unexpected symbol
 *	CFontTexturePage instead of undefined symbol TBmp in case neither Tga.h or Jpeg.h is included
 *
 *	2007-07-17
 *
 *	added line descent font parameter, ie. CFontTexturePage::m_n_char_descent, new field in
 *	CFontTexturePage constructor and CFontTexturePage::n_Line_descent()
 *	char descent is also added to bitmap info-block so it's not compatible with bitmaps
 *	created by the older versions
 *
 *	2007-12-24
 *
 *	improved linux compatibility by adding posix integer types
 *
 *	2008-03-04
 *
 *	now using Integer.h header
 *
 */

#ifndef __FONT_TEXTURE_INCLUDED
#define __FONT_TEXTURE_INCLUDED

#include "Integer.h"

#ifdef WIN32

#define FONT_TEXTURE_CREATE_FONTS
// only under win32, when defined, it's possible to create font pages on the fly using GDI
// (saves space taken by font textures, should be generaly faster than outline fonts drawn
// as polygons under OpenGL)

#endif //WIN32

#ifdef FONT_TEXTURE_CREATE_FONTS

/*
 *	class CGDI_Bitmap
 *		- simple class, enwrapping GDI bitmap for drawing it and onto it
 */
class CGDI_Bitmap {
protected:
	int m_n_width;
	int m_n_height;
	int m_n_plane_num;
	int m_n_bit_count;
	unsigned char *m_p_buffer;
	BITMAPINFO m_t_info;
	HBITMAP m_h_dib;

	HDC m_h_dc; // dc for rendering to bitmap
	HGDIOBJ m_h_prev_dc_obj;

public:
	/*
	 *	CGDI_Bitmap::CGDI_Bitmap(int n_width, int n_height, int n_plane_num = 1,
	 *		int n_bit_count = 32)
	 *		- default constructor. creates bitmap with size n_width per n_height pixels,
	 *		  n_plane_num color planes and n_bit_count bits per pixel
	 *		- note uncompressed RGB formats are supported only
	 */
	CGDI_Bitmap(int n_width, int n_height, int n_plane_num = 1, int n_bit_count = 32);

	/*
	 *	CGDI_Bitmap::~CGDI_Bitmap()
	 *		- default destructor
	 */
	~CGDI_Bitmap();

	/*
	 *	inline bool b_State() const
	 *		- in case constructor succeeded in creating bitmap, returns true. otherwise false
	 */
	inline bool b_State() const
	{
		return m_h_dib != 0;
	}

	/*
	 *	inline int n_Width() const
	 *		- returns bitmap width
	 */
	inline int n_Width() const { return m_n_width; }

	/*
	 *	inline int n_Height() const
	 *		- returns bitmap height
	 */
	inline int n_Height() const { return m_n_height; }

	/*
	 *	inline int n_Plane_Num() const
	 *		- returns number of bitmap color planes
	 */
	inline int n_Plane_Num() const { return m_n_plane_num; }

	/*
	 *	inline int n_Bit_Count() const
	 *		- returns number of bits per pixel
	 */
	inline int n_Bit_Count() const { return m_n_bit_count; }

	/*
	 *	inline const unsigned char *p_Buffer() const
	 *		- returns pointer to bitmap buffer (do not delete!)
	 */
	inline unsigned char *p_Buffer() { return m_p_buffer; }

	/*
	 *	inline const unsigned char *p_Buffer() const
	 *		- returns const pointer to bitmap buffer (do not try to delete!)
	 */
	inline const unsigned char *p_Buffer() const { return m_p_buffer; }

	/*
	 *	HDC CGDI_Bitmap::h_GetDC()
	 *		- creates device context and assigns this bitmap to be rendered to
	 */
	HDC h_GetDC();

	/*
	 *	bool CGDI_Bitmap::ReleaseDC()
	 *		- releases device context
	 *		- returns true on success or in case no context was created, otherwise false
	 */
	bool ReleaseDC();

	/*
	 *	void CGDI_Bitmap::Draw(HDC h_dc, int n_x, int n_y, int n_op = SRCCOPY) const
	 *		- draws bitmap using device context h_dc on position n_x, n_y
	 *		  using operation n_op
	 */
	void Draw(HDC h_dc, int n_x, int n_y, int n_op = SRCCOPY) const;
};

#endif //FONT_TEXTURE_CREATE_FONTS

/*
 *	struct TCharacterInfo
 *		- structure, holding info on character (metrics, bitmap, texture page placement)
 */
struct TCharacterInfo {
	unsigned char n_code; // character code

	int n_bitmap_width;
	int n_bitmap_height; // size of actual bitmap containing character (without offset)
	Vector2i v_baseline_offset; // offset from baseline point to bottom-left bitmap corner
	int n_caret_shift; // width of character base (position shift for the next character)

	Vector2i v_bitmap_pos; // origin in character atlas

#ifdef FONT_TEXTURE_CREATE_FONTS

	/*
	 *	bool TCharacterInfo::Create(unsigned char n_character, CGDI_Bitmap &r_bitmap)
	 *		- create character info using font selected in r_bitmap. n_character is requested
	 *		  character ascii code and r_bitmap serves as temporal buffer for drawing
	 *		  (must have sufficient size)
	 *		- returns true on success or false on failure (not enough memory)
	 */
	bool Create(unsigned char n_character, CGDI_Bitmap &r_bitmap);

	/*
	 *	void TCharacterInfo::Render(HDC h_dc, int n_raster_height) const
	 *		- renders character using device context h_dc of
	 *		- n_raster_height is texture page height (windows render upside-down)
	 */
	void Render(HDC h_dc, int n_raster_height) const;

	/*
	 *	inline unsigned int TCharacterInfo::n_Area(int n_padding) const
	 *		- returns character bitmap area, including padding
	 *		- useful for priorizing characters when inserting them to texture pages
	 *		  (big ones first, small ones will fill spaces)
	 */
	inline unsigned int n_Area(int n_padding) const
	{
		return (n_bitmap_width + n_padding) * (n_bitmap_height + n_padding);
	}

	/*
	 *	inline void TCharacterInfo::Place(const Vector2i &r_v_min)
	 *		- places bitmap to position r_v_min pixels in texture page
	 */
	inline void Place(const Vector2i &r_v_min)
	{
		v_bitmap_pos = r_v_min;
	}

	/*
	 *	inline void TCharacterInfo::Shift_Right(const TCharacterInfo &r_t_other, int n_padding)
	 *		- shifts bitmap right so it doesn't collide with r_t_other
	 */
	inline void Shift_Right(const TCharacterInfo &r_t_other, int n_padding)
	{
		v_bitmap_pos.x = r_t_other.v_bitmap_pos.x +
			r_t_other.n_bitmap_width + 1 + n_padding; // padding
	}

	/*
	 *	bool TCharacterInfo::b_Collide(const TCharacterInfo &r_t_other, int n_padding)
	 *		- returns true in case bitmaps of this character and r_t_other
	 *		  are colliding in texture space, otherwise false
	 */
	bool b_Collide(const TCharacterInfo &r_t_other, int n_padding);

#endif //FONT_TEXTURE_CREATE_FONTS

	/*
	 *	inline Vector2i TCharacterInfo::v_Min() const
	 *		- returns bottom-left bitmap corner position in texture space
	 */
	inline Vector2i v_Min() const
	{
		return v_bitmap_pos;
	}

	/*
	 *	inline Vector2i TCharacterInfo::v_Max(int n_padding) const
	 *		- returns top-right bitmap corner position in texture space
	 */
	inline Vector2i v_Max(int n_padding) const
	{
		return v_bitmap_pos + Vector2i(n_bitmap_width + n_padding,
			n_bitmap_height + n_padding); // padding
	}

protected:
#ifdef FONT_TEXTURE_CREATE_FONTS
	static inline bool b_Disjunct(int n_min1, int n_max1, int n_min2, int n_max2)
	{ return n_max1 < n_min2 || n_min1 > n_max2; }
	static bool b_Scanline_Empty(const uint32_t *p_buffer, int n_width);
	static bool b_Column_Empty(const uint32_t *p_buffer, int n_height, int n_width);
#endif //FONT_TEXTURE_CREATE_FONTS
};

/*
 *	class CFontTexturePage
 *		- texture page class
 */
class CFontTexturePage {
protected:
	std::vector<TCharacterInfo*> m_character_list;
	TCharacterInfo m_t_info_block;
	int m_n_max_char_num;
	int m_n_width, m_n_height;
	int m_n_space_width, m_n_newline_height, m_n_char_descent, m_n_internal_leading;
	int m_n_padding;

	TCharacterInfo *m_p_character;
	// contains linear block of characters in m_character list (if constructed from bitmap)

#ifdef FONT_TEXTURE_CREATE_FONTS

	class CGDIRenderBitmaps {
	protected:
		HDC m_h_dc;
		int m_n_raster_height;

	public:
		inline CGDIRenderBitmaps(HDC h_dc, int n_raster_height)
			:m_h_dc(h_dc), m_n_raster_height(n_raster_height)
		{}

		inline void operator ()(const TCharacterInfo *p_info) const
		{
			p_info->Render(m_h_dc, m_n_raster_height);
		}
	};

	class CSolveCollisions {
	protected:
		TCharacterInfo &m_r_t_rect;
		const int m_n_width, m_n_height;
		const int m_n_padding;
		int &m_r_n_min_max_y;

	public:
		CSolveCollisions(TCharacterInfo &r_t_rect, int n_width, int n_height,
			int n_padding, int &r_n_temp)
			:m_r_t_rect(r_t_rect), m_n_width(n_width), m_n_height(n_height),
			m_n_padding(n_padding), m_r_n_min_max_y(r_n_temp)
		{}

		inline bool operator ()(const TCharacterInfo *p_other)
		{
			if(m_r_t_rect.b_Collide(*p_other, m_n_padding)) {
				if(p_other->v_Max(m_n_padding).y < m_r_n_min_max_y)
					m_r_n_min_max_y = p_other->v_Max(m_n_padding).y + 1;
				// track minimum of max-y

				m_r_t_rect.Shift_Right(*p_other, m_n_padding);
				// shift right

				if(m_r_t_rect.v_Max(0).x >= m_n_width) {
					m_r_t_rect.Place(Vector2i(1, m_r_n_min_max_y));
					m_r_n_min_max_y = m_n_height;
				}
				// restart position

				return true;
			}
			return false;
		}
	};

	class CGrayBlockWriter {
	protected:
		unsigned char *m_p_block;
		const int m_n_scan_length;
		const int m_n_block_width;
		int m_n_scan_pos;

	public:
		inline CGrayBlockWriter(void *p_block, int n_scan_length, int n_block_width)
			:m_p_block((unsigned char*)p_block), m_n_scan_length(n_scan_length - n_block_width),
			m_n_block_width(n_block_width), m_n_scan_pos(n_block_width)
		{}

		inline void Write(const void *p_data, unsigned int n_size)
		{
			const unsigned char *p_byte_data = (const unsigned char*)p_data;
			while(1) {
				if(n_size < m_n_scan_pos) {
					m_n_scan_pos -= n_size;
					_copy(p_byte_data, n_size);
					return;
				} else if(n_size == m_n_scan_pos) {
					_copy(p_byte_data, n_size);
					m_p_block += (m_n_scan_length) * 4;
					m_n_scan_pos = m_n_block_width;
					return;
				} else /*if(n_size > m_n_scan_pos)*/ {
					_copy(p_byte_data, m_n_scan_pos);
					p_byte_data += m_n_scan_pos;
					n_size -= m_n_scan_pos;
					m_p_block += (m_n_scan_length) * 4;
					m_n_scan_pos = m_n_block_width;
				}
			}
		}

	protected:
		inline void _copy(const unsigned char *p_gray_src, unsigned int n_size)
		{
			for(const unsigned char *p_end = p_gray_src + n_size;
			   p_gray_src != p_end; ++ p_gray_src) {
				*m_p_block ++ = *p_gray_src;
				*m_p_block ++ = *p_gray_src;
				*m_p_block ++ = *p_gray_src;
				++ m_p_block; // don't care about alpha
			}
		}
	};

	template <class CBlockWriter>
	class CWriteIntLayout {
	protected:
		CBlockWriter m_writer;

	public:
		inline CWriteIntLayout(CBlockWriter writer)
			:m_writer(writer)
		{}

		inline void operator ()(const TCharacterInfo *p_char_info)
		{
			m_writer.Write(p_char_info, sizeof(TCharacterInfo));
		}
	};

	class CSumArea {
	protected:
		int m_n_area;
		int m_n_padding;

	public:
		inline CSumArea(int n_padding)
			:m_n_padding(n_padding), m_n_area(0)
		{}

		inline void operator ()(const TCharacterInfo &r_char_info)
		{
			m_n_area += r_char_info.n_Area(m_n_padding);
		}

		inline operator int() const
		{
			return m_n_area;
		}
	};

#endif //FONT_TEXTURE_CREATE_FONTS

	class CGrayBlockReader {
	protected:
		const unsigned char *m_p_block;
		const unsigned int m_n_scan_length;
		const unsigned int m_n_block_width;
		unsigned int m_n_scan_pos;

	public:
		inline CGrayBlockReader(const void *p_block, unsigned int n_scan_length,
			unsigned int n_block_width)
			:m_p_block((const unsigned char*)p_block), m_n_scan_length(-n_scan_length -
			n_block_width), m_n_block_width(n_block_width), m_n_scan_pos(n_block_width)
		{}

		inline void Read(void *p_data, unsigned int n_size)
		{
			unsigned char *p_byte_data = (unsigned char*)p_data;
			while(1) {
				if(n_size < m_n_scan_pos) {
					m_n_scan_pos -= n_size;
					_reverse_copy(p_byte_data, n_size);
					return;
				} else if(n_size == m_n_scan_pos) {
					_reverse_copy(p_byte_data, n_size);
					m_p_block += (m_n_scan_length) * 4;
					m_n_scan_pos = m_n_block_width;
					return;
				} else /*if(n_size > m_n_scan_pos)*/ {
					_reverse_copy(p_byte_data, m_n_scan_pos);
					p_byte_data += m_n_scan_pos;
					n_size -= m_n_scan_pos;
					m_p_block += (m_n_scan_length) * 4;
					m_n_scan_pos = m_n_block_width;
				}
			}
		}

	protected:
		inline void _reverse_copy(unsigned char *p_gray_dest, unsigned int n_size)
		{
			for(unsigned char *p_end = p_gray_dest + n_size; p_gray_dest != p_end;
			   ++ p_gray_dest, m_p_block += 4)
				*p_gray_dest = *m_p_block;
		}
	};

public:
	typedef std::vector<TCharacterInfo>::iterator tcharinfo_iterator;

#ifdef FONT_TEXTURE_CREATE_FONTS

	/*
	 *	CFontTexturePage::CFontTexturePage(tcharinfo_iterator p_character_begin,
	 *		tcharinfo_iterator p_character_end, int n_padding, int n_space_width,
	 *		int n_newline_height, int n_line_descent, int n_internal_leading, int n_max_size)
	 *		- default constructor; attempts to insert characters p_character_begin till
	 *		  p_character_end while keeping maximal texture dimensions below or equal
	 *		  to n_max_size while keeping 1 pixel spacing between characters and border and
	 *		  n_padding spacing between characters themselves
	 *		- n_space_width, n_newline_height, n_line_descent and n_internal_leading
	 *		  are font metrics (in pixels)
	 *		- note power-of-two texture dimensions are used only
	 *		- call n_Character_Num() to see how many of characters were actually added 
	 */
	CFontTexturePage(tcharinfo_iterator p_character_begin,
		tcharinfo_iterator p_character_end, int n_padding, int n_space_width,
		int n_newline_height, int n_line_descent, int n_internal_leading, int n_max_size);

	/*
	 *	tcharinfo_iterator CFontTexturePage::p_AddCharacters(tcharinfo_iterator p_character_begin,
	 *		tcharinfo_iterator p_character_end)
	 *		- attempts to insert characters p_character_begin till
	 *		  p_character_end while keeping texture dimensions set by constructor
	 *		- returns next character to be inserted to another texture page
	 */
	tcharinfo_iterator p_AddCharacters(tcharinfo_iterator p_character_begin,
		tcharinfo_iterator p_character_end);

#endif //FONT_TEXTURE_CREATE_FONTS

	/*
	 *	CFontTexturePage::CFontTexturePage(const TBmp *p_bitmap)
	 *		- loads texture page from bitmap p_bitmap
	 *		- assumes grayscale layout info stored in bottom-left corner of bitmap
	 *		- in case of success, n_Character_Num() should return positive number of characters
	 *		  (font tool should never create empty texture page), otherwise there will be
	 *		  no characters
	 */
	CFontTexturePage(const TBmp *p_bitmap);
	
	/*
	 *	CFontTexturePage::~CFontTexturePage()
	 *		- default destructor
	 */
	~CFontTexturePage()
	{
		if(m_p_character)
			delete[] m_p_character;
	}

	/*
	 *	inline int CFontTexturePage::n_Width() const
	 *		- return texture page width
	 */
	inline int n_Width() const
	{
		return m_n_width;
	}

	/*
	 *	inline int CFontTexturePage::n_Height() const
	 *		- return texture page height
	 */
	inline int n_Height() const
	{
		return m_n_height;
	}

	/*
	 *	inline int CFontTexturePage::n_Padding() const
	 *		- return character padding in pixels
	 */
	inline int n_Padding() const
	{
		return m_n_padding;
	}

	/*
	 *	inline int CFontTexturePage::n_Space_Width() const
	 *		- return space width (carret offset) in pixels
	 */
	inline int n_Space_Width() const
	{
		return m_n_space_width;
	}

	/*
	 *	inline int CFontTexturePage::n_Newline_Height() const
	 *		- return distance between two baseline one above the other, in pixels
	 */
	inline int n_Newline_Height() const
	{
		return m_n_newline_height;
	}

	/*
	 *	inline int CFontTexturePage::n_Line_Descent() const
	 *		- return space required under baseline
	 */
	inline int n_Line_Descent() const
	{
		return m_n_char_descent;
	}

	/*
	 *	inline int CFontTexturePage::n_Internal_Leading() const
	 *		- return space occupied by punctation marks (included in height)
	 */
	inline int n_Internal_Leading() const
	{
		return m_n_internal_leading;
	}

	/*
	 *	inline int CFontTexturePage::n_Max_Character_Num() const
	 *		- returns maximal number of characters that can be contained in this texture page
	 */
	inline int n_Max_Character_Num() const
	{
		return m_n_max_char_num;
	}

	/*
	 *	inline int CFontTexturePage::n_Character_Num() const
	 *		- returns number of characters contained in this texture page
	 */
	inline int n_Character_Num() const
	{
		if(!m_character_list.size())
			return 0;
		return m_character_list.size() - 1;
	}

	/*
	 *	inline const TCharacterInfo &CFontTexturePage::t_Character(int n_index) const
	 *		- return n_index-th character from this page
	 */
	inline const TCharacterInfo &t_Character(int n_index) const
	{
		return *m_character_list[n_index + 1];
	}

	/*
	 *	template <class CFunctionObject>
	 *	inline CFunctionObject CFontTexturePage::for_each_character(CFunctionObject t_object) const
	 *		- std::for_each lookalike for looping trough all characters
	 *		  contained in this texture page
	 */
	template <class CFunctionObject>
	inline CFunctionObject for_each_character(CFunctionObject t_object) const
	{
		return std::for_each(m_character_list.begin() + 1, m_character_list.end(), t_object);
	}

#ifdef FONT_TEXTURE_CREATE_FONTS

	/*
	 *	CGDI_Bitmap *CFontTexturePage::p_Create_Bitmap(HFONT h_font) const
	 *		- renders bitmap with glyphs (must be the same font as the one used
	 *		  for glyph rendering otherwise glyphs won't fit)
	 *		- returns true on success, false on failure
	 */
	CGDI_Bitmap *p_Create_Bitmap(HFONT h_font) const;

#endif //FONT_TEXTURE_CREATE_FONTS
};

#ifdef FONT_TEXTURE_CREATE_FONTS

class CFontTexture_Util {
public:
	/*
	 *	static bool CFontTexture_Util::Create_CharacterList(HDC h_dc, unsigned char n_first_char,
	 *		unsigned char n_char_num, std::vector<TCharacterInfo> &r_character_list,
	 *		CGDI_Bitmap &r_bitmap)
	 *		- fills r_character_list with list of characters in range n_first_char trough
	 *		  n_first_char + n_char_num
	 *		- r_bitmap is used for glyph rendering
	 *		- returns true on success, false on failure
	 */
	static bool Create_CharacterList(unsigned char n_first_char,
		unsigned char n_char_num, std::vector<TCharacterInfo> &r_character_list,
		CGDI_Bitmap &r_bitmap);

	/*
	 *	class CFontTexture_Util::CSortCharacters
	 *		- small function object for sorting characters descendingly by their areas with padding
	 *		  (better space use efficiency if big ones are inserted first)
	 */
	class CSortCharacters {
	protected:
		const int m_n_padding;

	public:
		inline CSortCharacters(int n_padding = 1)
			:m_n_padding(n_padding)
		{}

		inline bool operator ()(const TCharacterInfo &r_char_a,
			const TCharacterInfo &r_char_b) const
		{
			return r_char_a.n_Area(m_n_padding) > r_char_b.n_Area(m_n_padding);
		}
	};

	/*
	 *	static bool CFontTexture_Util::Create_TexturePageList(std::vector<CFontTexturePage*>
	 *		&r_texture_list, std::vector<TCharacterInfo> &r_character_list, int n_char_padding,
	 *		int n_max_texture_size, int n_space_width, int n_newline_height, int n_line_descent,
	 *		int n_internal_leading)
	 *		- fills list of texture pages r_texture_list by characters from r_character_list
	 *		  while keeping n_char_padding pixels distances between characters and
	 *		  having texture pages with dimensions equal or below n_max_texture_size
	 *		- n_space_width, n_newline_height, n_line_descent and n_internal_leading
	 *		  are font metrics
	 *		- returns true on success, false on failure
	 */
	static bool Create_TexturePageList(std::vector<CFontTexturePage*> &r_texture_list,
		std::vector<TCharacterInfo> &r_character_list, int n_char_padding,
		int n_max_texture_size, int n_space_width, int n_newline_height, int n_line_descent,
		int n_internal_leading);
};

#endif //FONT_TEXTURE_CREATE_FONTS

#endif //__FONT_TEXTURE_INCLUDED
