/*
								+---------------------------------+
								|                                 |
								|  ***   Tiny jpeg decoder   ***  |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|             Jpeg.h              |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 2006-08-30
 *
 *	write code for __JPEG_DEC_ALLOW_SUBSAMPLED_IMAGES code path (not really necessary)
 *	write code to support 12-bit extended DCT images
 *	refine code for lossless compression to eliminate rounding error when decompressing > 8bit
 *	lossless images (alloc just two int16 scanlines to remember predictor values)
 *
 *	2007-09-27
 *
 *	fixed error with restart interval being set to 0 (setting it to 0 means disable it
 *	which is rather strange, it would be preferable not to mention it at all)
 *
 *	2007-12-24
 *
 *	improved linux compatibility by adding posix integer types
 *
 *	2008-02-25
 *
 *	added function for loading jpegs from memory
 *
 *	2008-03-04
 *
 *	using Integer.h header, using CallStack.h instead of crtdbg.h
 *	renamed CJpegDecoder::CFileBuffer to CJpegDecoder::CBufferredFile
 *	changed size of zigZag table back to 64 and it's type to int (speed / obfuscation purposes)
 *
 *	2008-03-11
 *
 *	fixed error on line 1281 where macroblock index overrun check was "less than 63",
 *	it should've been 64 (error carried-in on 2008-03-04 update)
 *
 */

#ifndef __BMP_INCLUDED
#define __BMP_INCLUDED

#include "Integer.h"

struct TBmp {
	char n_former_bpp; // former bpp, before conversion to RGBA8
	bool b_grayscale;
	bool b_alpha;
	int n_width, n_height;
	uint32_t *p_buffer;
};

#endif //__BMP_INCLUDED

#ifndef __TINY_JPEG_INCLUDED
#define __TINY_JPEG_INCLUDED

/*
 *		=== config section ===
 */

#define __JPEG_DEC_MAX_COMPONENT_NUM 4
// maximal component number. decoding routines are written to support arbitrary number of
// components (colorspace conversion routines support 1 - 4 components), almost all
// implementatinos use 4, maximum possible component count in jpeg is 255

//#define __JPEG_DEC_VERBOSE
// if defined, prints tables, header info, ... (to stdout)

//#define __JPEG_DEC_READ_VERBOSE
// if defined, prints some debug info about file input (to stdout)

//#define __JPEG_DEC_DECODE_VERBOSE
// if defined, prints decoding progress (i.e. huffman codes, DCT coeffs; to stdout)

//#define __JPEG_DEC_ALLOW_SUBSAMPLED_GRAYSCALE
// jpeg specs allow grayscale images with sampling 1x1 only. if defined, decoder is able to
// render grayscale images with other sampling factors as well (do not affect rendering of
// color images) resulting code is larger when enabled

//#define __JPEG_DEC_SUPPORT_LOSSLESS
// if defined, decoder can handle lossless jpeg images as well

#ifdef __JPEG_DEC_SUPPORT_LOSSLESS
#define __JPEG_DEC_USE_LOSSLESS_NONSTD_RLE
#endif
// if defined, jpeg lossless algorithm is further enhanced by RLE compression this is non-standard
// extension which requires all images to be written by encoder, supporting this extension
// never use in programs where your jpeg files are available in standalone-file form so they could
// be potentialy copied away! do not breed nor allow anyone to breed non-standard jpegs away from
// your app (note there is no record in image file header wheter this extension was used or not)

#ifdef __JPEG_DEC_SUPPORT_LOSSLESS
//#define __JPEG_DEC_ALLOW_SUBSAMPLED_IMAGES
// do not use! there's no code to support it yet (todo - next version)
#endif
// if defined, decoder can decode lossless images with subsampling other than 1x1
// (it's a bit useless thing to bother with lossless compression and then downsample the
// image; moreover it's not very logical to do colorspace transformation in lossless mode
// as that would introduce more error into image so unless you store your images in YCbCr(K)
// format, subsampling is useless as well ... but despite all that, it's possible)

//#define __JPEG_DEC_SUPPORT_RG
// enable branch that can unpack RG images (i use them for uv-maps like water distortion, etc)
// another practice on how to interpret two-channel jpeg images is 16-bit RGB (acd-see)

//#define __JPEG_DEC_SUPPORT_YCCK
// enable branch that can unpack YCCK images

#ifndef __JPEG_DEC_SUPPORT_YCCK
//#define __JPEG_DEC_SUPPORT_CMYK
#endif
// enable branch that can unpack CMYK images (jpeg does contain component number only,
// not component meaning so it's possible to handle 4-component jpegs either as YCCK (common)
// or as CMYK (possible, but not widely used))

//#define __JPEG_DEC_INTEGER_IDCT
// if defined, use faster integer IDCT. otherwise use precise, slower float IDCT
// (in msvc, it's faster only in optimize speed mode. in optimize size, it's a little bit slower)

#if defined(__JPEG_DEC_INTEGER_IDCT) && !defined(_DEBUG)
#define __JPEG_DEC_USE_MMX
#endif
// if defined, faster MMX integer IDCT path is used, note code is a bit larger (~500 bytes)

//#define __JPEG_DEC_STRIP_ERROR_CHECKS
// if defined, all error checks aparts from alloc success tests are removed - saves a few code bytes
// (arround 1 kB; use with precaution - errors during decoding process may make your program crash)

#define __JPEG_DEC_BGR
// if defined, swap R and B during colorspace transformation

/*
 *		=== ~config section ===
 */

class CJpegDecoder {
protected:
	static int p_zig_indices[64]; // actualy inverse zig sequence

	class CBufferredFile { // class for fast buffered file input (also assures byte-order reversion)
	protected:
		unsigned char _m_p_buffer_[16384], *m_p_buffer;
		static const int m_n_buffer_size;
		int m_n_buffer_used;
		const unsigned char *m_p_buffer_ptr;
		FILE *m_p_fr;

	public:
		CBufferredFile(FILE *p_fr);
		CBufferredFile(const unsigned char *p_data, unsigned int n_size);
		bool Read_Byte(uint8_t &r_n_byte);
		bool Read_Short(int16_t &r_n_short);
		bool inline Read_Short(uint16_t &r_n_short) { return Read_Short((int16_t&)r_n_short); }
		bool Read_Marker(uint16_t &r_n_marker_code);
		bool Skip_Bytes(int n_byte_num);
	};

	class CBitReader { // class for reading bit-by-bit, used in huffman decoder
	protected:
		uint8_t m_n_cur_byte; // current byte, it's shifted as we eat bits
		bool m_b_0xff_byte; // set to true if last byte read was 0xff
		uint8_t m_n_bit_num; // number of valid bits in byte
		bool m_b_error_flag;

	public:
		CBitReader();
		bool inline b_Error() { return m_b_error_flag; }
		uint8_t n_GetBit(CBufferredFile *p_file_buffer);
		int16_t n_GetNumber(int n_bit_num, CBufferredFile *p_file_buffer);
		void Restart();
	};

	struct TBlockAction { // table for block actions
		uint16_t n_id;
		bool (*p_function)(uint16_t n_block_id, CJpegDecoder *p_this, CBufferredFile *p_file_buffer);
	};
	static const TBlockAction m_p_action_table[];

	struct TQuantTable { // quantization table
		bool b_16_bit; // otherwise 8-bit
#ifdef __JPEG_DEC_INTEGER_IDCT
		int p_value[64]; // 16:16 fixed-point (short would be enough for 8-bit quant tables, but not for 16 bit)
#else
		float p_value[64]; // hack! - should be unsigned short, but i pre-multiplied it and can use faster IDCT function
#endif
	} m_p_quant_table[4];
	// quantization tables (can be separate one for each component)

	struct THuffmanTable {
		bool b_valid; // did we read it from the file? in case we didn't, it could make our program crash, (unlikely if quantization tables are missing, image just looks scrambled but it can never make program crash)
		uint8_t p_code_num[16]; // number of codes for all possible 16 lengths
		uint8_t *p_code[16]; // 16 pointers into p_code_table where the actual data are
		uint16_t p_min_code_value[16]; // min and max code value; n+1 bit code must be >= p_min_code_value[n] && < p_max_code_value[n] (thus in case p_code_num[n] == 0 then also p_min_code_value[n] == p_max_code_value[n])
		//uint16_t p_max_code_value[16]; // don't store this, it's equal to p_min_code_value[n] + p_code_num[n]
		uint8_t p_code_table[256]; // table for all huffman codes
	} m_p_huffman_table[4][2]; // 0 = AC, 1 = DC
	// huffman tables (can be separate one for each component)

	struct TComponentInfo {
		uint8_t n_component_id; // 1 = Y, 2 = Cb, 3 = Cr (JFIF standard)
		uint8_t n_sampling_horiz;
		uint8_t n_sampling_vert; // component sampling (1, 2, or 4 usually)
		uint8_t n_scale_horiz;
		uint8_t n_scale_vert;
		uint8_t n_quant_table_index; // used quantization table
	};
	// outside TImageHeader because of scope issues (needed visibility for TScanHeader::TComponentInfo)

	struct TImageHeader {
		uint16_t n_frame_start_marker; // marker_SOF0_HuffBaseline_DCT supported for now only
		uint8_t n_sample_precision; // 8 (8 or 12 for extended baseline or progressive)
		int16_t n_height;
		int16_t n_width;
		int n_macro_block_num, n_macro_block_x_num; // number of macro-blocks
		uint8_t n_component_num; // should be 1-4 (1 = gray, 3 = YCbCr, 4 = CMYK)
		TComponentInfo p_component_info[__JPEG_DEC_MAX_COMPONENT_NUM];
	} m_t_image_header;
	// image header

	struct TRestartInterval {
		bool b_enabled;
		uint16_t n_interval;
	} m_t_restart;
	// restart interval definition

	struct TScanHeader {
		uint8_t n_component_num; // should match TImageHeader::n_component_num

		struct TComponentInfo2 {
			uint8_t n_component_id; // should match TImageHeader::TComponentInfo::n_component_id
			TComponentInfo *p_component; // points to image header component
			uint8_t n_DC_entropy_table;
			uint8_t n_AC_entropy_table; // table indices <0, 3>
			THuffmanTable *p_DC_table, *p_AC_table;
		} p_component_info[__JPEG_DEC_MAX_COMPONENT_NUM];

		uint8_t n_spectrum_first; // index (in zig-zag order) of first coded coefficient
		uint8_t n_spectrum_last; // index (in zig-zag order) of last coded coefficient (should be set to 63 for sequential DCT)
		uint8_t n_sucessive_approximation; // two nibbles, point transform, should be zero (not used in marker_SOF0_HuffBaseline_DCT)
	};

	struct TImageData {
		bool b_read_image_header;
#ifdef __JPEG_DEC_SUPPORT_LOSSLESS
		bool b_lossless; // set to true in case marker_SOF3_HuffLossless was read; valid only if b_read_image_header == true
#endif
		uint8_t *p_framebuffer[__JPEG_DEC_MAX_COMPONENT_NUM]; // frame-buffers, as big as image size / component sampling
		uint16_t p_framebuffer_width[__JPEG_DEC_MAX_COMPONENT_NUM];
		uint16_t p_framebuffer_height[__JPEG_DEC_MAX_COMPONENT_NUM];
		int16_t p_dc_value[__JPEG_DEC_MAX_COMPONENT_NUM]; // dc accumulators
	} m_t_image_data; // cleared always when starting decoding jpeg

	template <const int n_component_num, class TColorConversion>
	class CImageRenderer;
	// template body in Jpeg.cpp (don't want to put it here as it's pretty useless)

public:
	/*
	 *	TBmp *p_Decode_JPEG(FILE *p_fr)
	 *		- decode jpeg, convert it to RGBA, 8 bits per channel bitmap
	 *		- alpha is always 0xff since jpeg files don't store alpha channel
	 *		- return 0 on failure (not enough memory / unsupported format / damaged file)
	 */
	TBmp *p_Decode_JPEG(FILE *p_fr);

	/*
	 *	TBmp *CJpegDecoder::p_Decode_JPEG(const unsigned char *p_data, unsigned int n_size)
	 *		- decode jpeg, convert it to RGBA, 8 bits per channel bitmap
	 *		- alpha is always 0xff since jpeg files don't store alpha channel
	 *		- return 0 on failure (not enough memory / unsupported format / damaged file)
	 */
	TBmp *CJpegDecoder::p_Decode_JPEG(const unsigned char *p_data, unsigned int n_size);

	/*
	 *	void Supply_QuantTable(int n_index, const void *p_table, bool b_16_bit)
	 *		- it's possible to supply quantization tables before decoding image
	 *		- (image is decoded anyway, even if it doesn't contain quantization
	 *		  tables and you can save some space if all your images use the same
	 *		  quantization tables)
	 *		- n_index has to be in range 0-3 (there are four tables in jpeg)
	 *		- p_table points to table data (if b_16_bit == false, then it's array
	 *		  of 64 unsigned chars, if b_16_bit == true then it's array of 64
	 *		  unsigned shorts)
	 */
	bool Supply_QuantTable(int n_index, const void *p_table, bool b_16_bit);

protected:
	TBmp *_p_Decode_JPEG(CBufferredFile &file_buffer);
	static bool Read_SkipMarker(uint16_t n_block_id, CJpegDecoder *p_this, CBufferredFile *p_file_buffer);
	static bool Read_Unsupported(uint16_t n_block_id, CJpegDecoder *p_this, CBufferredFile *p_file_buffer);
	static bool Read_Skip(uint16_t n_block_id, CJpegDecoder *p_this, CBufferredFile *p_file_buffer);
	static bool Read_JFIF_Header(uint16_t n_block_id, CJpegDecoder *p_this, CBufferredFile *p_file_buffer);
	static bool Read_RestartInterval(uint16_t n_block_id, CJpegDecoder *p_this, CBufferredFile *p_file_buffer);
	static bool Read_QuantTables(uint16_t n_block_id, CJpegDecoder *p_this, CBufferredFile *p_file_buffer);
	static bool Read_HuffmanTables(uint16_t n_block_id, CJpegDecoder *p_this, CBufferredFile *p_file_buffer);
	static bool Read_FrameHeader(uint16_t n_block_id, CJpegDecoder *p_this, CBufferredFile *p_file_buffer);
	static bool Read_StartScan(uint16_t n_block_id, CJpegDecoder *p_this, CBufferredFile *p_file_buffer);
	bool Decode_Baseline_MacroBlock(int n_macro_block_index, const TScanHeader *p_scan_head,
		CBitReader &r_bit_reader, CBufferredFile *p_file_buffer);
#ifdef __JPEG_DEC_SUPPORT_LOSSLESS
	bool Decode_Lossless_Image(const TScanHeader *p_scan_head,
		CBitReader &r_bit_reader, CBufferredFile *p_file_buffer);
#endif
};

#endif //__TINY_JPEG_INCLUDED
