/*
								+--------------------------------+
								|                                |
								|  ***   LamePak2 classes   ***  |
								|                                |
								|  Copyright  -tHE SWINe- 2008  |
								|                                |
								|           LamePak2.h           |
								|                                |
								+--------------------------------+
*/

/*
 *	2007-02-25
 *
 *	this is first beta version if the file. todo - store path separator in pak header,
 *	re write TBuffer so it can realloc itself in a way std::vector can and try to avoid
 *	any (re)allocations while unpacking files. add function to purge block cache so
 *	it doesn't keep memory when no files are unpacked for longer periods of time
 *
 *	2007-02-28
 *
 *	path separator is now in pak header and there's new packer verbose object.
 *	added compression method field to pak header.
 *
 */

#ifndef __LAME_PAK_2_INCLUDED
#define __LAME_PAK_2_INCLUDED

#include "Integer.h"

/*
	pak file structure:

  TPak2Header
  TPak2FileInfo-0
  .
  .
  TPak2FileInfo-n
  block-0-size
  Block-0
  .
  .
  block-m-size
  Block-m
  0

	block sizes are 32 bit integers, in case MSB is raised,
	corresponding block is stored uncompressed.
 */

#pragma pack(1)
// tight alignemnt

/*
 *	struct TPak2Header
 *		- contains pak2 file header
 *		- note the structure is tightly aligned and can be written directly
 */
struct TPak2Header {
	int16_t n_version;					// major * 100 + minor
	int16_t n_method;					// packing method
	uint32_t n_file_num;				// number of files
	char n_path_separator;				// backslash character, used in paths
	uint32_t n_block_size;				// size of uncompressed blocks
	uint32_t n_block_num;				// number of blocks
	uint32_t n_toc_size;				// size of toc block (compressed or with raised MSB, as with block sizes)
};

/*
 *	enum <unnamed>
 *		- packing method names
 */
enum {
	method_BWT_A = 0,						// BWT packing method A (block order is BWT, MTF, RLE, Huff)
	method_BWT_B							// BWT packing method B (block order is BWT, MTF-1, RLE, Huff)
};

/*
 *	struct TPak2FileInfo
 *		- contains archived file information
 *		- note the structure is tightly aligned and can be written directly
 */
struct TPak2FileInfo {
	bool b_same_dir;					// file is in the same dir as the last one
	uint32_t n_size_lo, n_size_hi;		// file size
	uint32_t n_block, n_block_offset;	// position in the blocks
	uint16_t n_filename_length;			// filename length in bytes; terminating zero is not saved
};

#pragma pack()
// default alignment

#define PAK2_PACKER_USE_STDERR
// enable CPak2PackerUtil reporting it's errors to stderr

/*
 *	class CPak2PackerUtil
 *		- simple class containing function for creating pak from list of files
 */
class CPak2PackerUtil {
public:
	/*
	 *	class CPak2PackerUtil::CVerboseObject
	 *		- virtual class for progress indicator
	 *		- gets (differential) amounts of packer input and output data
	 *		- can be used to interrupt packing
	 */
	class CVerboseObject {
	public:
		/*
		 *	virtual void On_TOCCompress(unsigned int n_in, unsigned int n_out) = 0
		 *		- called when table of contents is compressed
		 *		- n_in is amount of data to be packed and n_out is resulting packed size
		 */
		virtual void On_TOCCompress(unsigned int n_in, unsigned int n_out) = 0;

		/*
		 *	virtual bool On_BlockCompress(unsigned int n_in, unsigned int n_out) = 0
		 *		- called when data block is compressed
		 *		- n_in is amount of data to be packed and n_out is resulting packed size
		 *		- returning false in this function makes packer fail immediately
		 */
		virtual bool On_BlockCompress(unsigned int n_in, unsigned int n_out) = 0;
	};

protected:
	class CBlockCounter;

	/*
	 *	TNamedPak2FileInfo
	 *		- TPak2FileInfo with associated filename
	 */
	typedef std::pair<TPak2FileInfo, const char*> TNamedPak2FileInfo;

	static const uint32_t n_raw_data_flag; // MSB bit
	static const char n_path_separator; // backslash
	static const short n_pak_version; // pak version

	unsigned int m_n_block_size;
	int m_n_compression_level;
	CVerboseObject *m_p_verbose_obj;
	int m_n_thread_num;

	TBuffer m_t_tmp, m_t_tmp2, m_t_tmp3;
	// temporary buffers for compression routine (to minimize allocation count)

public:
	/*
	 *	CPak2PackerUtil::CPak2PackerUtil()
	 *		- default constructor, sets default packer options
	 */
	CPak2PackerUtil();

	/*
	 *	unsigned int CPak2PackerUtil::n_BlockSize() const
	 *		- returns block size
	 */
	unsigned int n_BlockSize() const;

	/*
	 *	bool CPak2PackerUtil::Set_BlockSize(unsigned int n_block_size)
	 *		- sets new block size, valid values are 1k - 100M
	 *		- returns true for valid values, false for invalid values
	 *		  (invalid values aren't set)
	 */
	bool Set_BlockSize(unsigned int n_block_size);

	/*
	 *	int CPak2PackerUtil::n_CompressionLevel() const
	 *		- returns compression level option (the only valid compression
	 *		  levels at the time are 0: store and 1: compress (default))
	 */
	int n_CompressionLevel() const;

	/*
	 *	bool CPak2PackerUtil::Set_CompressionLevel(int n_compression_level)
	 *		- sets compression level, the only valid compression
	 *		  levels at the time are 0: store and 1: compress (default)
	 *		- returns true upon accepting valid value,
	 *		  false when rejecting given n_compression_level
	 */
	bool Set_CompressionLevel(int n_compression_level);

	/*
	 *	CVerboseObject *CPak2PackerUtil::p_VerboseObject() const
	 *		- returns current verbose object
	 */
	CVerboseObject *p_VerboseObject() const;

	/*
	 *	void CPak2PackerUtil::Set_VerboseFunc(CVerboseObject *p_verbose)
	 *		- p_verbose is verbose object which gets (differential) amounts of packer input
	 *		  and output data; can be 0
	 */
	void Set_VerboseFunc(CVerboseObject *p_verbose);

	/*
	 *	int CPak2PackerUtil::n_Thread_Num() const
	 *		- certain parts of compressor can run in parallel
	 *		- returns number of threads
	 */
	int n_Thread_Num() const;

	/*
	 *	bool CPak2PackerUtil::Set_Thread_Num(int n_thread_num)
	 *		- certain parts of compressor can run in parallel
	 *		- sets number of threads, valid values are 1 to 64
	 *		- returns true upon accepting valid value,
	 *		  false when rejecting given n_thread_num
	 */
	bool Set_Thread_Num(int n_thread_num);

	/*
	 *	static int CPak2PackerUtil::n_Version()
	 *		- returns packer version, which is stored as major * 100 + minor
	 */
	static int n_Version();

	/*
	 *	bool CPak2PackerUtil::CreatePak(const char *p_s_filename,
	 *		const std::vector<TFileInfo*> &r_file_list, const char *p_s_base_dir = 0)
	 *		- creates archive in p_s_filename
	 *		- r_file_list is list of input files, n_block_size is compression
	 *		  block size (couple of megabytes is ideal for compression, but
	 *		  smaller blocks compress faster)
	 *		- p_s_base_dir is base directory, path to it will not be included in pak file names
	 *		- returns true on success, false on failure
	 *		- note if PAK2_PACKER_USE_STDERR is defined, error messages are printed to stderr
	 */
	bool CreatePak(const char *p_s_filename,
		const std::vector<TFileInfo*> &r_file_list, const char *p_s_base_dir = 0);

protected:
	bool PackBuffer(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer);
	bool PackFiles(const std::vector<TFileInfo*> &r_file_list, FILE *p_fw);
	bool PackWriteBlock(const TBuffer &r_t_block, FILE *p_fw);
	static int n_TOC_Size(const std::vector<TNamedPak2FileInfo> &r_file_info);
	static bool Assemble_TOC(const std::vector<TNamedPak2FileInfo> &r_file_info,
		TBuffer &r_t_toc);
	static const char *p_s_SameDirName(const char *p_s_path, const char *p_s_prev_path);
	static TNamedPak2FileInfo t_FileInfo(const char *p_s_path, const char *p_s_prev_path,
		const CBlockCounter &r_block_counter, const TFileInfo &r_t_file_info);
	bool CreateFileHeaders(const std::vector<TFileInfo*> &r_file_list,
		const char *p_s_base_dir, std::vector<TNamedPak2FileInfo> &r_file_info,
		unsigned int &r_n_block_num);
};

/*
 *	class CLamePak2
 *		- class encapsulating pak filesystem and unpacker
 */
class CLamePak2 {
protected:
	typedef std::pair<TPak2FileInfo, char*> TNamedPak2FileInfo;

	FILE *m_p_fr;
	TPak2Header m_t_header;
	std::vector<TNamedPak2FileInfo> m_toc;

	static const uint32_t n_raw_data_flag; // MSB bit
	static const short n_pak_version; // supported pak version

	TBuffer m_t_cur_block;
	unsigned int m_n_cur_block;

	TBuffer m_t_tmp, m_t_tmp2;

public:
	/*
	 *	CLamePak2::CLamePak2(const char *p_s_filename)
	 *		- opens lame pak p_s_filename and reads table of contents
	 *		- check b_Status() to see if it was succesful
	 */
	CLamePak2(const char *p_s_filename);

	/*
	 *	CLamePak2::~CLamePak2()
	 *		- destructor
	 */
	~CLamePak2();

	/*
	 *	bool CLamePak2::b_Status() const
	 *		- returns true if pak file was opened succesfuly, otherwise returns false
	 */
	bool b_Status() const;

	/*
	 *	int CLamePak2::n_Pak_Version() const
	 *		- returns pak file version or 0 in case pak failed to open
	 */
	int n_Pak_Version() const;

	/*
	 *	char CLamePak2::n_Path_Separator() const
	 *		- returns used path separator character (ie. backslash)
	 *		  or 0 in case pak failed to open
	 */
	char n_Path_Separator() const;

	/*
	 *	int CLamePak2::n_File_Num() const
	 *		- returns number of files or 0 in case pak file was not opened succesfuly
	 */
	int n_File_Num() const;

	/*
	 *	const char *CLamePak2::p_s_FileName(int n_index) const
	 *		- returns filename of file with zero-based index n_index
	 */
	const char *p_s_FileName(int n_index) const;

	/*
	 *	uint64_t CLamePak2::n_FileSize(int n_index) const
	 *		- returns size of file with zero-based index n_index
	 */
	uint64_t n_FileSize(int n_index) const;

	/*
	 *	bool CLamePak2::UnpackFile(int n_index, TBuffer &r_t_file_data)
	 *		- unpacks file with zero-based index n_index, outputs to r_t_file_data
	 *		  (doesn't have to be allocated)
	 *		- returns true on success, false on failure
	 */
	bool UnpackFile(int n_index, TBuffer &r_t_file_data);

protected:
	bool Read_TOC();
	bool Seek_Block(unsigned int n_block);
	bool Fetch_Block(unsigned int n_block);
	bool UnpackBuffer(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer);
	static inline void DeleteName(TNamedPak2FileInfo &r_t_file_info);
};

#endif //__LAME_PAK_2_INCLUDED
