/*
								+--------------------------------+
								|                                |
								|  ***   Simple threading   ***  |
								|                                |
								|  Copyright  -tHE SWINe- 2008  |
								|                                |
								|           Thread.cpp           |
								|                                |
								+--------------------------------+
*/

#include <vector>
#include "../berLame_src/CallStack.h"
#include "Thread.h"

#ifdef WIN32

/*
 *								=== CThread (win32) ===
 */

CThread::CEmptyRunnable CThread::empty_runnable;

/*
 *	CThread::CThread()
 *		- default constructor, attaches empty runnable object to a thread
 */
CThread::CThread()
	:m_p_runnable(&empty_runnable), m_h_thread(0)
{}

/*
 *	CThread::CThread(CRunnable &r_runnable)
 *		- constructor, attaches runnable object to a thread
 */
CThread::CThread(CRunnable &r_runnable)
	:m_p_runnable(&r_runnable), m_h_thread(0)
{}

/*
 *	CThread::~CThread()
 *		- destructor
 */
CThread::~CThread()
{
	if(m_h_thread)
		CloseHandle(m_h_thread);
}

/*
 *	bool CThread::AttachRunnable(CRunnable &r_runnable)
 *		- attaches different runnable, thread must not be running
 *		- returns true on success, false on failure
 */
bool CThread::AttachRunnable(CRunnable &r_runnable)
{
	if(b_IsRunning())
		return false;
	m_p_runnable = &r_runnable;
	return true;
}

/*
 *	CRunnable &CThread::r_Runnable()
 *		- returns reference to runnable object
 */
CRunnable &CThread::r_Runnable()
{
	return *m_p_runnable;
}

/*
 *	const CRunnable &CThread::r_Runnable() const
 *		- returns const reference to runnable object
 */
const CRunnable &CThread::r_Runnable() const
{
	return *m_p_runnable;
}

/*
 *	bool CThread::Start()
 *		- starts the thread and returns immediately
 *		- note this fails in case the thread is already running
 *		- returns true on success, false on failure
 */
bool CThread::Start()
{
	if(b_IsRunning())
		return false;
	if(!(m_h_thread = CreateThread(0, 0, _run, this, 0, 0)))
		return false;
	return true;
}

/*
 *	bool CThread::Run()
 *		- starts the thread and waits untill it finishes
 *		- note this fails in case the thread is already running
 *		- also note this is not thread-safe (the Thread object doesn't change
 *		  it's state to running while in this function and also can't be stopped)
 *		- returns true on success, false on failure
 */
bool CThread::Run()
{
#ifdef __THREAD_SAFE_RUN
	if(!Start())
		return false;
	return Stop();
	// this is thread safe, but requires thread
#else //__THREAD_SAFE_RUN
	if(b_IsRunning())
		return false;
	m_p_runnable->Run();
	return true;
	// this simply invokes thread's runnable
#endif //__THREAD_SAFE_RUN
}

/*
 *	bool CThread::b_IsRunning()
 *		- returns true if thread is running, otherwise returns false
 */
bool CThread::b_IsRunning()
{
	if(!m_h_thread)
		return false;
	if(!WaitForSingleObject(m_h_thread, 0)) {
		// in case thread is not running anymore,
		// WaitForSingleObject will return 0

		CloseHandle(m_h_thread);
		m_h_thread = 0;
		// close thread handle
		return false;
	}
	return true;
}

/*
 *	bool CThread::Stop(bool b_force_kill = false)
 *		- waits for the thread to end, if b_force_kill is set, it's ended forcibly
 *		- returns true if thread is stopped, returns false if thread didn't
 *		  stop in given time or on failure
 */
bool CThread::Stop(bool b_force_kill)
{
	if(!b_IsRunning())
		return true;
	// not running anymore

	if(!b_force_kill) {
		if(WaitForSingleObject(m_h_thread, INFINITE))
			return false;
	} else {
		if(!TerminateThread(m_h_thread, 0))
			return false;
		// kill thread
	}
	// either waits or terminates

	CloseHandle(m_h_thread);
	m_h_thread = 0;
	// destroy thread handle

	return true;
}

/*
 *	static int CThread::n_CPU_Num();
 *		- utility function; returns number of CPU's
 */
int CThread::n_CPU_Num()
{
	SYSTEM_INFO t_sys_info;
	GetSystemInfo(&t_sys_info);
	return t_sys_info.dwNumberOfProcessors;
}

/*
 *	static unsigned long __stdcall CThread::_run(void *p_arg)
 *		- run function for WinApi; runs the thread and returns 0
 */
unsigned long __stdcall CThread::_run(void *p_arg)
{
	CThread *p_this = (CThread*)p_arg;
	// this

	p_this->m_p_runnable->Run();
	// run the thread

	return 0;
}

/*
 *								=== ~CThread (win32) ===
 */

#else //WIN32

/*
 *								=== CThread (linux) ===
 */

CThread::CEmptyRunnable CThread::empty_runnable;

/*
 *	CThread::CThread()
 *		- default constructor, attaches empty runnable object to a thread
 */
CThread::CThread()
	:m_p_runnable(&empty_runnable), m_b_running(false)
{
	pthread_mutex_init(&t_running_mutex, NULL);
}

/*
 *	CThread::CThread(CRunnable &r_runnable)
 *		- constructor, attaches runnable object to a thread
 */
CThread::CThread(CRunnable &r_runnable)
	:m_p_runnable(&r_runnable), m_b_running(false)
{
	pthread_mutex_init(&t_running_mutex, NULL);
}

/*
 *	CThread::~CThread()
 *		- destructor
 */
CThread::~CThread()
{
	pthread_mutex_destroy(&t_running_mutex);
}

/*
 *	bool CThread::AttachRunnable(CRunnable &r_runnable)
 *		- attaches different runnable, thread must not be running
 *		- returns true on success, false on failure
 */
bool CThread::AttachRunnable(CRunnable &r_runnable)
{
	if(pthread_mutex_lock(&t_running_mutex))
		return false;
	// lock mutex, we will change m_b_running

	if(m_b_running) {
		pthread_mutex_unlock(&t_running_mutex);
		return false;
	}
	// thread is running, can't attach other runable

	m_p_runnable = &r_runnable;

	pthread_mutex_unlock(&t_running_mutex);
	// unlock

	return true;
}

/*
 *	CRunnable &CThread::r_Runnable()
 *		- returns reference to runnable object
 */
CRunnable &CThread::r_Runnable()
{
	return *m_p_runnable;
}

/*
 *	const CRunnable &CThread::r_Runnable() const
 *		- returns const reference to runnable object
 */
const CRunnable &CThread::r_Runnable() const
{
	return *m_p_runnable;
}

/*
 *	bool CThread::Start()
 *		- starts the thread and returns immediately
 *		- note this fails in case the thread is already running
 *		- returns true on success, false on failure
 */
bool CThread::Start()
{
	if(pthread_mutex_lock(&t_running_mutex))
		return false;
	// lock mutex, we will change m_b_running

	if(m_b_running) {
		pthread_mutex_unlock(&t_running_mutex);
		return false;
	}
	// thread is running, can't start

	pthread_attr_t t_attr;
	pthread_attr_init(&t_attr);
	pthread_attr_setdetachstate(&t_attr, PTHREAD_CREATE_JOINABLE);
	if(pthread_create(&m_t_thread, &t_attr, _run, (void*)this)) {
		pthread_attr_destroy(&t_attr);
		pthread_mutex_unlock(&t_running_mutex); // don't forget!
		return false;
	}
	pthread_attr_destroy(&t_attr);
	// create joinable thread

	m_b_running = true;
	pthread_mutex_unlock(&t_running_mutex);
	// set running to true and unlock

	return true;
}

/*
 *	bool CThread::Run()
 *		- starts the thread and waits untill it finishes
 *		- note this fails in case the thread is already running
 *		- also note this is not thread-safe (the Thread object doesn't change
 *		  it's state to running while in this function and also can't be stopped)
 *		- returns true on success, false on failure
 */
bool CThread::Run()
{
	if(pthread_mutex_lock(&t_running_mutex))
		return false;
	// lock mutex, we will change m_b_running

	if(m_b_running) {
		pthread_mutex_unlock(&t_running_mutex);
		return false;
	}
	// thread is running, can't start

	m_b_running = true;
	pthread_mutex_unlock(&t_running_mutex);
	// set running to true and unlock

	m_p_runnable->Run();
	// this simply invokes thread's runnable

	if(pthread_mutex_lock(&t_running_mutex))
		_ASSERTE(0); // this shouldn't fail
	m_b_running = false;
	pthread_mutex_unlock(&t_running_mutex);
	// not running anymore

	return true;
}

/*
 *	bool CThread::b_IsRunning()
 *		- returns true if thread is running, otherwise returns false
 */
bool CThread::b_IsRunning()
{
	return m_b_running;
}

/*
 *	bool CThread::Stop(bool b_force_kill = false)
 *		- waits for the thread to end, if b_force_kill is set, it's ended forcibly
 *		- returns true if thread is stopped, returns false if thread didn't
 *		  stop in given time or on failure
 */
bool CThread::Stop(bool b_force_kill)
{
	if(pthread_mutex_lock(&t_running_mutex))
		return false;
	// lock mutex, we will change m_b_running

	if(!m_b_running) {
		pthread_mutex_unlock(&t_running_mutex);
		return true;
	}
	// not running anymore

	if(!b_force_kill) {
		pthread_mutex_unlock(&t_running_mutex);
		// unlock mutex, the thread will set m_b_running by itself

		if(pthread_join(m_t_thread, NULL))
			return false;
		// join the thread

		// don't assert m_b_running now! it can be running again by now!
	} else {
		if(pthread_cancel(m_t_thread)) {
			pthread_mutex_unlock(&t_running_mutex); // don't forget
			return false;
		}
		// kill thread

		m_b_running = false;
		pthread_mutex_unlock(&t_running_mutex);
		// set m_b_running 'manually'
	}
	// either waits or terminates

	return true;
}

#include <unistd.h>
#include <sys/sysctl.h>

#if defined(CTL_HW) && defined(HW_NCPU)

static int n_CPU_Num() // BSD
{
	int ncpus;
	size_t len = sizeof(ncpus);
	int mib[2] = {CTL_HW, HW_NCPU};
	if(sysctl(&mib[0], 2, &ncpus, &len, NULL, 0) != 0)
		return -1;
	return ncpus;
}

#elif defined(_SC_NPROCESSORS_ONLN)

static int n_CPU_Num() // Solaris
{
	return sysconf(_SC_NPROCESSORS_ONLN);
}

#elif defined(_SC_NPROCESSORS_CONF)

static int n_CPU_Num() // Linux
{
	return sysconf(_SC_NPROCESSORS_CONF);
}

#else

static int n_CPU_Num() // failed
{
	return -1;
}

#endif

/*
 *	static int CThread::n_CPU_Num();
 *		- utility function; returns number of CPU's
 *		- in case CPU count cannot be determined, returns -1
 */
int CThread::n_CPU_Num()
{
	return ::n_CPU_Num();
}

/*
 *	static void *CThread::_run(void *p_arg)
 *		- run function for pthread; runs the thread and returns 0
 */
void *CThread::_run(void *p_arg)
{
	CThread *p_this = (CThread*)p_arg;
	// this

	p_this->m_p_runnable->Run();
	// run the thread

	if(pthread_mutex_lock(&p_this->t_running_mutex))
		_ASSERTE(0); // this should't fail ... there's no way how to handle failure here
	p_this->m_b_running = false;
	pthread_mutex_unlock(&p_this->t_running_mutex);
	// not running anymore

	pthread_exit(NULL);
	// exit
}

/*
 *								=== ~CThread (linux) ===
 */

#endif //WIN32
